#! /usr/bin/perl

#
# omirc.pl - summarize irc activity
#
# may be run:
#	per sample
#	per N samples (e.g., all irc sample files in a daily log directory)
#
# For syntax:
#	# omirc.pl -d <directory> | -f "rawirc.txt0 ... rawirc.txtN"
#	    [ -html ] [ -debug ] > stdout
#
# at this time -html does nothing.
#
# input file(s):
#	irc.txt raw input file.  Note this file has two tuples in it.
#
# outputs:	   
#	stdout
#
# tuples used:  
#	IRCHOSTS
#	IRCCHAN
#
# author/s:
#	derek@cs.pdx.edu
#	jrb@cs.pdx.edu
#
# BUGS:
#
#	1. some of the channel names are dubious
#	2. might merge channel strings (which are not case sensitive) if and only if
#		their iplists and say one stat are the same.  However the problem with
#		this idea is that one muddies the information about how channels
#		are used ... and this may not be obvious to the ngrep user.
#

########################################################


use vars qw/ %opt /;
use Getopt::Long;

#my $opt_string = 'hwn:s:d:f:';
my %hosts = ();
my $directory="";
my $TRUE = 1;
my $FALSE = 0;
my $DEBUG = $FALSE;
my $HTML = $FALSE;
my $delimiter = "\n";

my $wormThreshold = 60;

if ( (scalar @ARGV) < 1 ) { 
	printUsageAndExit(); 
}

GetOptions("h" => \$opt{help},
	"html" => \$opt{html},
	"debug" => \$opt{debug},
	"n:i" => \$opt{n},
	"s:s" => \$opt{s},
	"d:s" => \$opt{d},
	"f:s" => \$opt{f},
);

if ( $opt{'h'} ) {
	printUsageAndExit();
}
if ( $opt{'d'} ) {
	$directory = $opt{'d'};

	#
	# read all files in a directory into the array and
	# prepend directory path to filename
	#
	opendir (DIR, $directory) or 
	    die "can't open $directory directory: $!\n";
	@FILES = readdir DIR;
	closedir DIR;
	$ninstances = $size = scalar @FILES;
	while ($size > -1) {
		# prepend directory path to files
		if ( -f "$directory/$FILES[$size]" ) {
			$FILES[$size] = $directory.'/'.$FILES[$size];
		}
		$size--;
	}
} 
if ( $opt{'f'} ) {
	@FILES = split / /, $opt{'f'};
	$ninstances = scalar @FILES;
}
#if ( $opt{'html'} ) {
#	$HTML = $TRUE;
#}
if ( $opt{'debug'} ) {
	$DEBUG = $TRUE;
}

# files are now in FILES array
# loop thru them
#
my $mode = 0;

while ( $_ = shift @FILES ) {

	open(FILE, $_) || die "can't open $_ file: $!\n";
	@FILE = <FILE>;
	close(FILE);

	$phrases = join("<br>",@FILE);

	#  fetch each line from datafile, line is separated by "\n"

	@phrases = split(/$delimiter/,$phrases);

	# walk phrases and process on line by line basis
	# one subroutine per filter type

	foreach my $elem (@phrases) {

		# get rid off blanks, return, line breaks

		$elem =~ s/\n//g;
		$elem =~ s/<br>//g;
		$elem =~ s/ //g;	

		# save each item to array @items, from each line,

		@items = split(/:/,$elem);

		# means a comment or a meta-command
		# either in host tuple mode or channel tuple mode
	 	if ($items[0] eq "#IRCNODE") {
			$mode = 1;
		    	next;
	 	}
	 	elsif ($items[0] eq "#IRCCHANNEL") {
			$mode = 2;
		    	next;
	 	}
		elsif($items[0] =~ /^#./){
		    	next;
		}

		#ipsrc:totalmsg:joins:pings:pongs:privmsgs:
		#joinsrc:joindst:pingsrc:pingdst:pongsrc:pongdst:
		#pmsrc:pmdst:channels:serverhits:
		#131.252.208.55:98:0:9:9:80:0:0:0:9:9:0:75:5:1:80:
		if ($mode == 1) {
#			print "1: $items[0]\n";
			$ipsrc = $items[0];
			$ircHostTotalMsg{$ipsrc} += $items[1];
			$ircHostJoins{$ipsrc} += $items[2];
			$ircHostPings{$ipsrc} += $items[3];
			$ircHostPongs{$ipsrc} += $items[4];
			$ircHostPrivMsgs{$ipsrc} += $items[5];
			# ignore other fields inbetween for now
			# calculate max channels for an ipsrc	
			$ircHostChannels{$ipsrc} += $items[14];
			if ($items[14] > $ircHostMaxChannels{$ipsrc}) {
				$ircHostMaxChannels{$ipsrc} = $items[14];
			}
			$ircHostServerHits{$ipsrc} += $items[15];
			# keep worst-case value.  
			# adding means we compute the average.
			$ircHostWorminess{$ipsrc} += $items[16];
			if ($items[16] > $ircHostMaxWorm{$ipsrc}) {
				$ircHostMaxWorm{$ipsrc} = $items[16];
			}
		}
		#channame:hits:joins:privmsgs:noips:ip list, ip, ip ...
		elsif ($mode == 2) {
#			print "2: $items[0]\n";
			$channel = $items[0];
			$ircChanHits{$channel} += $items[1];
			$ircChanJoins{$channel} += $items[2];
			$ircChanPrivmsgs{$channel} += $items[3];
			$ircChanIPCount{$channel} += $items[4];
			# $chanIPList{$channel} is kept up by this function
			ipChanListMerge($items[5], $channel);
		}
	}  # end loop
} # end file loop

globalStats();
chanStats();
hostStats();

exit;


########################################################## 

#
# subroutines
#

#
# global stats
#
#	# instances
#	# total number of irc hosts
#	# total number of irc servers
#	# total number of unique channels
#
# TBD:  later - biggest network?
#	average network size is likely to be 2
 
sub globalStats
{
	my $c = 0;
	my $start_time=`date`;
	$start_time=~ s/\n//g;

	print "irc summarization at: $start_time\n";

	print "###################################################\nglobal stats:\n";
	print "\t# of sample files: $ninstances\n";
	my $irchosts = keys %ircHostTotalMsg;
	print "\t# of irc hosts (servers and non-servers): $irchosts\n";
	foreach my $ipsrc (keys %ircHostServerHits) {
		if ( $ircHostServerHits{$ipsrc}) {
			$c++;
		}
	}
	print "\t# of irc servers: $c\n";
	my $uniquechannels =  keys %ircChanHits;
	print "\t# of unique channels: $uniquechannels\n";
}

#
# sort by biggest number of messages
#
# list channels
#	give per channel stats
#
# list channels again but with hosts in channel
#	list hosts in channel with host stats
#
# list subset of channels that are join only
#
sub chanStats
{
	my @wormychannels = undef;
	my $evilchannel = 0;

	print "\n\n###################################################\nchannel stats:\n\n";

	# sort the channels by max messages and list them
	#
	my @Allsortedchans  = sort chanhits_by_value keys(%ircChanHits);

	# now figure out how many worms per channel
	# this is the "evil" channel calculation
	# this calculation has one trick.  
	# note: no printing here!  this is a preprocess operation.
	# 
	foreach my $channel (@Allsortedchans) {
		$evilchannel = "";
		# first get the host count and host list
		my @iplist = split(",", $chanIPList{$channel});
		my $ipcount = @iplist;
		my $wormcount = 0;
		my $servercount = 0;
		foreach $host (@iplist) {
			# get the worm count for this host
			my $maxworm = $ircHostMaxWorm{$host};
			if ($maxworm >= $wormThreshold) {
				$wormcount++;
			}
			if ( $ircHostServerHits{$host} ) {
				$servercount++;
			}
		}
		my $totalhosts = $ipcount - $servercount;
		# now we have total ip count
		#             total worm count
		#	      total server count
		#	      total host count
		#
		# if half of the total hosts are wormy
		# if half of the total hosts discounting servers are wormy
		if ( $wormcount > 0 && (($wormcount >= ($ipcount / 2)) || ($wormcount >= (totalhosts / 2)))) {
			$evilchannel = E;
			if ($wormcount == 1) {
				$evilchannel = e;
			}
		}
		#if ( $wormcount > 0 && ($wormcount >= ($totalhosts / 2))) {
		#	$evilchannel = E;
		#}
		$chanEviltude{$channel} = $evilchannel;
		$chanWormCount{$channel} = $wormcount;
	}

	my @Allevilchans  = sort chan_by_evil keys(%chanWormCount);
	print "channels sorted by evil factor: max number of wormy hosts:\n";
	print "\tchannel                              msgs      joins   privmsgs    ipcount wormyhosts      evil?\n";
	foreach my $channel (@Allevilchans) {
		if ($chanWormCount{$channel} == 0) {
			last;
		}
		my @iplist = split(",", $chanIPList{$channel});
		my $ipcount = @iplist;
		my $sbuf = sprintf("%-30.30s %10d %10d %10d %10d %10d      %s",
			$channel, $ircChanHits{$channel}, $ircChanJoins{$channel}, $ircChanPrivmsgs{$channel}, $ipcount, $chanWormCount{$channel}, $chanEviltude{$channel});
		print "\t$sbuf\n";
	}

	print "\nchannels sorted by max messages (note e/E for possible evil channel):\n";
	print "\tchannel                              msgs      joins   privmsgs    ipcount wormyhosts      evil?\n";
	foreach my $channel (@Allsortedchans) {
		my @iplist = split(",", $chanIPList{$channel});
		my $ipcount = @iplist;
		my $sbuf = sprintf("%-30.30s %10d %10d %10d %10d %10d      %s",
			$channel, $ircChanHits{$channel}, $ircChanJoins{$channel}, $ircChanPrivmsgs{$channel}, $ipcount, $chanWormCount{$channel}, $chanEviltude{$channel});
		print "\t$sbuf\n";
	}

	# now list the channels with host stats
	print "\nchannels with per host stats:\n";
	print "\tchannel\n\t\tip_src                 tmsg  tjoin  tping  tpong tprivmsg maxchans maxworm Server?\n";
	foreach $channel (@Allsortedchans) {
		$sbuf = sprintf("%-30.30s", $channel);
		print "\t$sbuf\n";

		my @iplist = split(",", $chanIPList{$channel});
		my $newchan = 0;
		foreach $host (@iplist) {
			# are we a server or a host
			if ( $ircHostServerHits{$host} ) {
				$sflag = "S";
			}
			else {
				$sflag = "H";
			}
			my $tmsg = $ircHostTotalMsg{$host}; 
			my $tjoin = $ircHostJoins{$host};
			my $tping = $ircHostPings{$host};
			my $tpong = $ircHostPongs{$host};
			my $tprivmsg = $ircHostPrivMsgs{$host}; 
			my $maxchan = $ircHostMaxChannels{$host};
			my $maxworm = $ircHostMaxWorm{$host};
			if (($maxworm >= $wormThreshold) && ($newchan == 0)) {
				push @wormychannels, $channel;
				$newchan = 1;
			}
			$sbuf = sprintf("%-20.20s %6d %6d %6d %6d %8d %8d %6d %6s",
				$host, $tmsg, $tjoin, $tping, $tpong, $tprivmsg, $maxchan, $maxworm, $sflag);
			print "\t\t$sbuf\n";
		}
	}

	# list channels that only have join messages
	print "\nchannels with no privmsgs, only joins:\n";
	print "\tchannel                              msgs      joins   privmsgs    ipcount\n";
	foreach $channel (@Allsortedchans) {
		if ($ircChanJoins{$channel} == $ircChanHits{$channel}) {
			my @iplist = split(",", $chanIPList{$channel});
			my $ipcount = @iplist;
			my $sbuf = sprintf("%-30.30s %10d %10d %10d %10d",
				$channel, $ircChanHits{$channel}, $ircChanJoins{$channel}, $ircChanPrivmsgs{$channel}, $ipcount);
			print "\t$sbuf\n";
		}
	}

	# list channels that seem to be wormy where we define wormy 
	# as any channel with one host greater than the wormThreshold
	print "\nchannels with ANY hosts with non-zero possibly wormy work metric (> $wormThreshold%):\n";
	print "\tchannel";
	foreach $channel (@wormychannels) {
		print "\t$channel\n";
	}
}

#
# hostStats
#
# list servers - busy to not busy
# TBD: list "our" servers
#
# list subset of hosts that are join only
#
# list subset of hosts that have any AVERAGE worminess
#

sub hostStats
{
	print "\n\n###################################################\nirc host stats:\n";
	# list servers
	#
	my @Allsortedhosts  = sort hosthits_by_value keys(%ircHostTotalMsg);

	print "servers sorted by max messages\n";
	print "\tip_src                 tmsg  tjoin  tping  tpong tprivmsg maxchans maxworm\n";
	foreach my $host (@Allsortedhosts) {
		if ( $ircHostServerHits{$host} ) {
			my $tmsg = $ircHostTotalMsg{$host}; 
			my $tjoin = $ircHostJoins{$host};
			my $tping = $ircHostPings{$host};
			my $tpong = $ircHostPongs{$host};
			my $tprivmsg = $ircHostPrivMsgs{$host}; 
			my $maxchan = $ircHostMaxChannels{$host};
			my $maxworm = $ircHostMaxWorm{$host};
			$sbuf = sprintf("%-20.20s %6d %6d %6d %6d %8d %8d %6d",
				$host, $tmsg, $tjoin, $tping, $tpong, $tprivmsg, $maxchan, $maxworm);
			print "\t$sbuf\n";
		}
	}

	# list hosts that are join only with no privmsg
	print "\nhosts with join msgs but no privmsgs\n";
	print "\tip_src                 tmsg  tjoin  tping  tpong tprivmsg maxchans maxworm\n";
	foreach my $host (@Allsortedhosts) {
		my $tjoin = $ircHostJoins{$host};
		my $tprivmsg = $ircHostPrivMsgs{$host}; 

		if (($tprivmsg == 0) && ($tjoin > 0)) {
			my $tmsg = $ircHostTotalMsg{$host}; 
			my $tjoin = $ircHostJoins{$host};
			my $tping = $ircHostPings{$host};
			my $tpong = $ircHostPongs{$host};
			my $tprivmsg = $ircHostPrivMsgs{$host}; 
			my $maxchan = $ircHostMaxChannels{$host};
			my $maxworm = $ircHostMaxWorm{$host};
			$sbuf = sprintf("%-20.20s %6d %6d %6d %6d %8d %8d %6d",
				$host, $tmsg, $tjoin, $tping, $tpong, $tprivmsg, $maxchan, $maxworm);
			print "\t$sbuf\n";
		}
	}

	# list hosts that have shown any signs of worminess
	print "\nhosts with any signs of worminess with any instance of work >= $wormThreshold%\n";
	print "\tip_src                 tmsg  tjoin  tping  tpong tprivmsg maxchans maxworm\n";
	foreach my $host (@Allsortedhosts) {

		my $maxworm = $ircHostMaxWorm{$host};
		if ($maxworm >= $wormThreshold) {
			my $tmsg = $ircHostTotalMsg{$host}; 
			my $tjoin = $ircHostJoins{$host};
			my $tping = $ircHostPings{$host};
			my $tpong = $ircHostPongs{$host};
			my $tprivmsg = $ircHostPrivMsgs{$host}; 
			my $maxchan = $ircHostMaxChannels{$host};
			$sbuf = sprintf("%-20.20s %6d %6d %6d %6d %8d %8d %6d",
				$host, $tmsg, $tjoin, $tping, $tpong, $tprivmsg, $maxchan, $maxworm);
			print "\t$sbuf\n";
		}
	}
}


sub printUsageAndExit()
{
	print "Usage: $0 -d directory -f \"file1 file2 ... fileN\"\n";
	print "     -h            displays usage\n";
        print "     -d directory  directory of mon.lite files\n";
        print "     -f file1 file2 ... fileN list of files to read\n";
        print "     -html         produce HTML output\n";
        print "     -debug        produce debugging columns\n";
	exit 1;
}

#
# merge ip list into one unique list
#
# global: 
#	$chanIPList{$channel}
sub ipChanListMerge
{
	my $newips = $_[0];
	my $channel = $_[1];

	my @newiplist = split(",", $newips);
	my @oldiplist = split(",", $chanIPList{$channel});

	my $last = undef;
	my @new = grep { ($last ne $_) && ($last = $_) } sort(@newiplist, @oldiplist);

	$chanIPList{$channel} = join(",", @new);

#	print "$channel: $chanIPList{$channel}\n";
}

sub chanhits_by_value 
{
        if ($ircChanHits{$a} < $ircChanHits{$b}) {
                1;
        }  
        elsif ($ircChanHits{$a} == $ircChanHits{$b}) {
                0;
        }
        elsif ($ircChanHits{$a} > $ircChanHits{$b}) {
                -1;
        }
}

sub chan_by_evil 
{
        if ($chanWormCount{$a} < $chanWormCount{$b}) {
                1;
        }  
        elsif ($chanWormCount{$a} == $chanWormCount{$b}) {
                0;
        }
        elsif ($chanWormCount{$a} > $chanWormCount{$b}) {
                -1;
        }
}

sub hosthits_by_value 
{
        if ($ircHostTotalMsg{$a} < $ircHostTotalMsg{$b}) {
                1;
        }  
        elsif ($ircHostTotalMsg{$a} == $ircHostTotalMsg{$b}) {
                0;
        }
        elsif ($ircHostTotalMsg{$a} > $ircHostTotalMsg{$b}) {
                -1;
        }
}
